/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.filters.PostFilter;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.IntermediateDataSet;
import de.bsvrz.pua.prot.processing.util.TempAttribut;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingInformation.ApplyAggregation;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.pua.prot.util.attributes.*;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.NoChangeMarker;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Tuple;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;

/**
 * Klasse zum Markieren und Erstellen von Ausgangsdatensätzen. Zum Starten der Datenaufbereitung den Thread mit {@link Thread#start()} starten. Wenn Datensätze
 * eingehen, wird die Datenaufbereitung benachrichtigt. Ist die Datensammlung abgeschlossen, wird die Datenaufbereitung ein letzes Mal benachrichtigt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class ProcessingBuffer extends Thread {
	/*
	 * OPT Markierungsinformation nur pro Datenidentifikation ablegen & verwalten -> genauso wie 'winner'
	 */

	/** Initialisierung Ringpuffer */
	protected static final int INITIAL_RINGBUFFER_SIZE = 5000;

	/**
	 * Soll ein leerer Datensatz eingefügt werden? Wird in {@link #getWinners(ValueProvider[],List)} gesetzt, und in {@link #run()} verwendet. Der Wert in dieser
	 * Varoable wird laufen überschrieben, daher muss sie innerhalb der Schleife in {@link #run()} gesichert werden.
	 */
	private boolean _insertEmpty = false;

	/** Informationen zur Datenaufbereitung */
	protected ProcessingInformation _processingInformation;

	/** Objekt das beim Eintreffen von Daten benachrichtigt werden soll. */
	private ProcessingInterface _processor = null;

	/** Informationen über das Ergebnis der Datensammlung */
	protected ProtocolRequestResult _bufferResult = new ProtocolRequestResult();

	/** Verbindung zum Dateverteiler */
	protected ClientDavInterface _connection;

	/** Debug Ausgaben. Nicht static, da der Logger für jeden neuen Thread neu angelegt werden soll! */
	protected final Debug _debug = Debug.getLogger();

	/** Ringpuffer in dem die Ausgangsdatensätze abgelegt werden. Ausgangsdatensätze sind für den Versand vorbereitete Daten. */
	// XXX MH Zugriff auf den Buffer Prüfen: Wird der überhaupt noch gebraucht, oder kann man die Ergebnisse dem
	// Processing sofort rüberschieben? Wo wird er noch lokal aufgerufen?
	protected RingBuffer<IntermediateDataSet> _buffer;

	/** Zuordnung: Reale Elemente -&gt; Wertebeschaffer */
	protected Hashtable<RealElement, ValueProvider> _realElements;

	/** Zuordnung: Temporäre Attribute -&gt; Wertebeschaffer für temporäre Attribute */
	protected Hashtable<TempAttributeDescription, TempAttribut> _tempElements;

	/**
	 * Abzufragende Zeitbereiche. Einträge dürfen nach der Initialisierung nicht mehr gelöscht werden (und sollten nicht verändert werden, daher wird die Liste im
	 * Konstruktor unveränderbar gemacht.
	 */
	protected final List<Tuple<Long, Long>> _periods;

	/** Zeigt an ob aggregiert werden soll. */
	private boolean _aggregate = false;

	/** Zeigt an ob alle Daten gesammelt und abgerufen wurden. */
	protected boolean _done;

	/** Zeigt ob die Datenaufbereitung abgebrochen wurde. */
	private boolean _abort = false;

	/** Objekt dass die Ausgangsdatensätze erzeugt. Wird in {@link #run()} initialisiert. */
	protected IMDSBuilder _dataSetBuilder = null;

	private final int _startPeriodIndex;
	private final String _scriptName;

	/**
	 * Startet den Buffer. {@link #init()} sollte unverzüglich nach Erzeugen des Objekts aufgerufen werden.
	 *
	 * @param connection       Verbindung zum Datenverteiler
	 * @param processor Objekt das die Datenaufbereitung durchführt. Wird jedes Mal benachrichtigt, wenn ein Ausgangsdatensatz vorliegt.
	 * @param processingInformation        Informationen zur Datenaufbereitung.
	 * @param periods   Zeitbereiche, in denen der Buffer Daten sammeln soll. Die Zeitbereiche müssen bereits sortiert und zusammengefasst sein! Die Liste wird
	 *                  nicht geändert.
	 *
	 * @param startPeriodIndex StartPeriodIndex
	 * @param scriptName Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration
	 */
	public ProcessingBuffer(
			final ClientDavInterface connection,
			final ProcessingInterface processor,
			final ProcessingInformation processingInformation,
			final List<Tuple<Long, Long>> periods,
			final int startPeriodIndex,
			final String scriptName)
			throws FailureException {
		this._connection = connection;
		this._processor = processor;
		this._processingInformation = processingInformation;
		_startPeriodIndex = startPeriodIndex;
		_scriptName = scriptName;
		this._periods = Collections.unmodifiableList(periods);
		_buffer = new RingBuffer<IntermediateDataSet>(INITIAL_RINGBUFFER_SIZE, RingBuffer.UNBOUNDED_SIZE);

		_realElements = new Hashtable<RealElement, ValueProvider>();
		_tempElements = new Hashtable<TempAttributeDescription, TempAttribut>();

		this.setName(this.getClass().getName() + ": " + processor.getJobId());
	}

	/**
	 * Erzeugt die Abbildungen: Elemente -&gt; Wertebeschaffer. Meldet sich für jede benötigten Datenidentifikation beim Datenverteiler an. Füllt die Hashtables
	 * {@link #_realElements} und {@link #_tempElements}. Die Objekte, die in diesen Hashtables als 'Values' abgelegt werden, erledigen das Abfragen der Werte
	 * selbständig. Sollte sofort nach Erzeugen des Objektes Aufgerufen werden.
	 *
	 * @throws FailureException     Fehler.
	 * @throws InterruptedException Thread wird abgebrochen.
	 */
	public abstract void init() throws FailureException, InterruptedException;

	/** Legt einen Eintrag für jedes temporären Attribut in {@link #_tempElements} an. */
	protected void setTempAttributes() {
		// Temporäre Attribute
		for(final Elements element : _processingInformation.getAttributes()) {
			if(element instanceof TempAttributeDescription) {
				final TempAttributeDescription tad = (TempAttributeDescription)element;
				if(_tempElements.get(tad) == null) setTimeStampOrigin(tad, _processingInformation.getMainObject());
			}
		}
	}

	/**
	 * Speichert eine Referenz auf das Objekt, dessen Zeitstempel von diesem temporären Attribut verwendet werden soll. Sortiert alle dabei gefundenen temporären
	 * Attribute in {@link #_tempElements} ein.
	 *
	 * @param tad        Temporäres Objekt, dessen Zeitstempel gewonnen werden soll.
	 * @param mainObject Hauptobjekt des Protokolls.
	 */
	private RealAttributeDescription setTimeStampOrigin(final TempAttributeDescription tad, final SystemObject mainObject) {
		/* Verwendet NICHT die getTimeStampOriginMethode von TempAttributeDescription, da diese nicht so performat ist. */
		final TempAttribut tempAtt = _tempElements.get(tad);
		if(tempAtt != null) {
			return tempAtt.getTimeStampOrigin();
		}
		else {
			RealAttributeDescription timeStampOrigin = null;
			RealAttributeDescription tempAttOrigin = null;

			for(final AttributeDescription ad : tad.getExpression().getAttributes()) {
				if(ad instanceof RealAttributeDescription) {
					final RealAttributeDescription rad = (RealAttributeDescription)ad;
					if(rad.qualifier.object.equals(mainObject)) // Attribut das das Hauptobjekt verwendet
					{
						timeStampOrigin = rad;
						break;
					}
					else if(timeStampOrigin == null) // Erstes Reales Attribut
					{
						timeStampOrigin = rad;
					}
				}
				else if(timeStampOrigin == null) // temporäre Attribute nur durchsuchen, wenn es keine realen gibt.
				{
					if(ad instanceof TempAttributeDescription) {
						tempAttOrigin = setTimeStampOrigin((TempAttributeDescription)ad, mainObject);
					}
					else if(ad instanceof DurationAttributeDescription) {
						/*
						 * Erstes Attribut, dass das Haupt-Objekt verwendet als Origin verwenden.
						 */
						for(final RealAttributeDescription rad : _processingInformation.getAllRealAttributes()) {
							if(rad.qualifier.object.equals(mainObject)) {
								timeStampOrigin = rad;
								break;
							}
						}
					}
				}
			}
			timeStampOrigin = (timeStampOrigin != null) ? timeStampOrigin : tempAttOrigin;
			_tempElements.put(tad, new TempAttribut(tad, timeStampOrigin, _processingInformation.getProtocolType()));
			return timeStampOrigin;
		}
	}

	/**
	 * Zeigt ob die Datensammlung erfolgreich war.
	 *
	 * @return Objekt mit Informationen über den Ablauf der Datensammlung
	 */
	public ProtocolRequestResult getResult() {
		return _bufferResult;
	}

	/** Benachrichtigt das Objekt das die Datenaufbereitung durchführt. */
	protected void notifyProcessor() {
		synchronized(_processor) {
			_processor.notifyAll();
		}
	}

	/**
	 * Liefert einen Ausgangsdatensatz zurück. Setzt dabei das 'zeitdauer' Attribut. Liegt keiner vor, kehrt die Methode sofort mit dem Rückgabewert null zurück.
	 *
	 * @return Ausgangsdatensatz und Status oder null, falls keiner vorliegt.
	 *
	 * @throws InterruptedException Warten auf Ausgangsdatensatz wurde unterbrochen.
	 */
	public IntermediateDataSet take() throws InterruptedException {
		if(hasData()) {
			try {
				if(_dataSetBuilder.isDone() && _buffer.size() == 1 && isDoneCollecting()) setDone();
			}
			catch(FailureException e) {
				_bufferResult = new ProtocolRequestResult(e.getMessage());
				setDone();
			}

			return _buffer.pop();
		}
		else {
			return null;
		}
	}

	/**
	 * Erstellt die Ausgangsdatensätze. Fängt sämtliche(!) Exceptions ab. Informationen zum Ablauf der Datensammlung können mittels {@link
	 * ProcessingBuffer#getResult()} abgefragt werden.
	 */
	@Override
	public void run() {
		try {
			int i = 0;

			final boolean merge;
			if(_processingInformation.getConstraint() != null && _processingInformation.getConstraint().last != null) {
				merge = _processingInformation.getConstraint().last;
			}
			else {
				merge = false;
			}
			_dataSetBuilder = new IMDSBuilder(
					merge,
					_processingInformation.getNoChangeMarker() == NoChangeMarker.Cell,
					_processingInformation.getAttributes(),
					_processingInformation.getConstraint(),
					_realElements,
					_tempElements,
					this,
					new Status(_periods, _processingInformation.getPositionsToIgnoreForStatus()),
					_startPeriodIndex
			);

			// Array mit allen realen Elementen
			ValueProvider elements[] = new ValueProvider[_realElements.size()];
			elements = _realElements.values().toArray(elements);

			// winners enthält alle realen Elemente, für die Daten abgeholt werden müssen
			final ArrayList<ValueProvider> winners = new ArrayList<ValueProvider>();

			while(!isDoneCollecting()) {
				i++;
				if(i == 5000) {
					/*
					 * Auf dem Hauptsystem 'verhungern' manchen Threads fast, weil aus z.B. 5 Threads 2 bevorzugt
					 * ausgeführt werden -> yield
					 */
					Thread.yield();
					// debug.info("Performance Info: ProcessingBuffer (bisher gesendet: "+i+" -> DS/s: "+
					// i/((System.currentTimeMillis()-timestamp)/1000.0));
					i = 0;
				}

				processNewData(elements, winners);
			}
			/*
			 * letzten Datensatz senden Datenaufbereiter hier erneut aufwecken, wenn Sammlung der Daten abgeschlossen
			 * (-> Onlinebuffer weckt DA auf wenn fertig, Buffer weckt DA auch auf wenn es nur genau EINEN
			 * Ausgangsdatensatz gibt!)
			 */
			_dataSetBuilder.createIntermediateDataSet(true);
			notifyProcessor();
			// Ende des Zeitbereiches erreicht: Datensammeln beenden.
			for(final ValueProvider vp : _realElements.values()) {
				vp.abort();
			}
		}
		catch(FailureException e) {
			setDone();
			e.log();
			this._bufferResult = new ProtocolRequestResult(e.getMessage());
		}
		catch(InterruptedException e) // Thread wird abgebrochen -> kein Fehler!
		{
			setDone();
			_debug.finest("Thread " + this.getName() + " wird abgebrochen.");
		}
		catch(Exception e) // Alle übrigen Exceptions, z.B. zuwenig Speicher etc. werden hier abgefangen
		{
			setDone();
			_debug.error(e.getMessage(), e);
			this._bufferResult = new ProtocolRequestResult(Util.getStackTrace(e));
		}
		finally {
			_debug.fine(this.getClass() + " ( " + this.getName() + " ) hat die Datensammlung abgeschlossen.");
			// Abschließend noch einmal die DA benachrichtigen um Abschluss der Datensammlung mitzuteilen
			notifyProcessor();
		}
	}

	/**
	 * Bereitet aus den Ergebnisdatensätzen die Ergebnisdaten auf.
	 *
	 * @param elements Liste aller Datenlieferanten.
	 * @param winners  Rückgabe: Liste der Datensätze mit minimalem, nicht aufgefüllten Datenzeitstempel. (Es sollte eine leere ArrayList übergeben werden)
	 *
	 * @throws FailureException     Fehler bei der Aufbereitung.
	 * @throws InterruptedException Aufbereitung wurde unterbrochen.
	 */
	protected abstract void processNewData(ValueProvider[] elements, ArrayList<ValueProvider> winners) throws FailureException, InterruptedException;

	/**
	 * Werden nicht aggregierte Daten versendet?
	 *
	 * @return <code>true</code> falls <code>Liste</code> eine der ausgewählten die Aggregationsanwendungen ist.
	 */
	protected abstract boolean isListAggregation();

	/**
	 * Speichert den Ausgangsdatensatz im Ausgangspuffer {@link #_buffer}. Benachrichtigt den Datenaufbereiter {@link #_processor} falls notwendig.
	 *
	 * @param intermediateData Ausgangsdatensatz.
	 */
	public void storeDataSet(final IntermediateDataSet intermediateData) {
		_buffer.push(intermediateData);

		/*
		 * Datenaufbereiter erst benachrichtigen, wenn mehr als zwei Ausgangsdatensätze vorliegen. (Sonst könnte die
		 * Markierungsinformation noch verändert werden)
		 */
		if(_buffer.size() >= 2) notifyProcessor();
	}

	/**
	 * Führt die Aggregierungen durch. Die Aggregierungen werden jedoch nur durchgeführt, falls {@link #isAggregate()} <code>true</code> liefert.
	 *
	 * @param baseData Werte des Ausgangsdatensatz. Einträge können von der Methode geändert werden.
	 *
	 * @return <code>true</code>: Der Ausgangsdatensatz soll nachgefiltert und ausgegeben werden.
	 */
	public abstract boolean applyAggregations(List<BaseDataSet> baseData);

	/**
	 * Falls Aggregationsdatensätze vorhanden sind, werden sie in den Ausgangspuffer gelegt.
	 *
	 * @param status Status, den der Aggregationsdatensatz erhalten soll.
	 *
	 * @see #storeDataSet(IntermediateDataSet)
	 */
	public abstract void storeAggregatedData(byte status);

	/**
	 * Ruft den Nachfilter für nicht-aggregierte Daten auf.
	 *
	 * @param baseData        Elemente des Ausgangsdatensatz.
	 * @param status          Status, der dem Ausgangsdatensatz zugeordnet werden soll.
	 * @param winnerTimeStamp Minimaler, nicht aufgefüllter Zeitstempel.
	 * @param currentInterval CurrentInterval
	 *
	 * @return Durch den Nachfilter bearbeiteter Datensatz.
	 */
	public IntermediateDataSet applyPostFilter(final List<BaseDataSet> baseData, final byte status, final long winnerTimeStamp, final int currentInterval) {
		final IntermediateDataSet ids = new IntermediateDataSet(baseData, status, ApplyAggregation.LISTE, winnerTimeStamp, currentInterval);
		final PostFilter postFilter = _processingInformation.getUserFilter();
		if(postFilter != null) {
			return postFilter.processDataRecord(ids);
		}
		else {
			return ids;
		}
	}

	/**
	 * Legt die Elemente mit dem kleinsten Datenzeitstempel in einer Liste ab. Setzt teilt den Elementen mit, ob sie zu den "Gewinnern" oder "Verlierern" gehören.
	 *
	 * @param elements Feld mit realen Elementen, deren Datenzeitstempel überprüft wird.
	 * @param winners  Liste in der die Elemente mit minimalem Datenzeitstempel abgelegt werden. Die Liste wird vorher geleert.
	 *
	 * @return Wert größer 0, falls es ein Element aus dem Spaltenbereich den kleinsten Datenzeitstempel hat. Wert kleiner 0, falls ein Element, das nicht im
	 *         Spalten-Bereich vorkommt, den kleinsten Datenzeitstempel hat. 0 wird zurückgegeben, wenn kein Element den minimalen Zeitstempel hat.
	 */
	protected int getWinners(final ValueProvider[] elements, final List<ValueProvider> winners) {
		_insertEmpty = false;
		// OPT Verschränken: Elemente gleich sortiert nach DZS ablegen!
		winners.clear();
		long minTimeStamp = Long.MAX_VALUE;
		// 1. Elemente mit dem kleinsten DZS suchen
		final ArrayList<ValueProvider> temps = new ArrayList<ValueProvider>();
		for(final ValueProvider element : elements) {
			// OPT als Verlierer markieren
			element.lost();
			/*
			 * nur als Gewinner markieren, falls für dieses Element noch weitere Daten zu erwarten sind. (Wenn nicht,
			 * dann ergibt sich eine Endlosschleife ergeben: Es werden ja nur für die Winner-Elemente neue Daten
			 * abgeholt. -> Nur durch das neue Daten abholen wird hasNext beeinflusst -> es werden keine neuen Daten
			 * abgeholt, "winner" bleibt ewig der "Gewinner".
			 */
			if(minTimeStamp >= element.getMostRecentTimeStamp() && element.hasNext())

			{
				temps.add(element);
				minTimeStamp = element.getMostRecentTimeStamp();
			}
		}
		// Elemente mit kleinestem Datenzeitstempel als "Gewinner" kennzeichnen.
		int gotWinner = 0;
		for(final ValueProvider element : temps) {
			if(element.getMostRecentTimeStamp() == minTimeStamp) {
				element.won();
				winners.add(element);
				if(element.isShown()) {
					gotWinner = 1;
				}
				else if(gotWinner == 0){
					gotWinner = -1;
				}
				/*
				 * es sollen nur die Gewinner Einfluß darauf haben, ob ein leerer Datensatz eingefügt wird. Andernfalls
				 * würde dieser leere Datensatz evtl "zu früh" eingfügt. Bsp: temporäres Attribut, das mehrere reale
				 * Attribute referenziert. Es liegen zwar schon Daten für alle Attribute vor (A mit DZS 100, B mit DZS
				 * 200), aber die Werte der realen Attribute werden der Reihe nach abgearbeitet: D.h. es wird der Wert
				 * für die DZ 100 berechnet -> es soll noch kein leerer Datensatz eingefügt werden!
				 */
				_insertEmpty |= element.dataInformation.insertEmpty();
			}
		}
		return gotWinner;
	}

	/** @return Anzahl Ausgangsdatensätze. */
	public int size() {
		return _buffer.size();
	}

	/**
	 * Zeigt an ob mit das Abfragen aller Daten abgeschlossen wurde. D.h. es wird <code>true</code> zurückgegeben, sobald alle Daten abgefragt wurden. {@link
	 * #isDone()} muss jedoch noch nicht <code>false</code> zurückliefern! Wird von {@link #run()} jedesmal aufgerufen, bevor ein neuer Ausgangsdatensatz erstellt
	 * wird.
	 *
	 * @return <code>true</code> falls alle Daten abgefragt wurden.
	 *
	 * @throws FailureException     Fehler bei der Überprüfung
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	protected abstract boolean isDoneCollecting() throws FailureException, InterruptedException;

	/**
	 * Zeigt an ob Daten abgeholt werden können.
	 *
	 * @return <code>true</code> falls Daten mittels {@link #take()} abgeholt werden können.
	 */
	public abstract boolean hasData();

	/**
	 * Wird aufgerufen wenn die Datensammlung beendet werden soll ist oder abgebrochen wird. Nach dem Aufruf dieser Methode <b>muss</b> {@link #_done}
	 * <code>true</code> zurückliefern.
	 */
	protected void setDone() {
		_done = true;
	}

	/**
	 * Zeigt ob der Buffer noch weitere Daten liefern wird.
	 *
	 * @return <code>true</code> falls noch weitere Daten zu erwarten sind.
	 */
	abstract public boolean isDone();

	/** Beendet die Datensammlung. Ruft dazu {@link ValueProvider#abort()} für jedes reale Element auf. */
	public void abort() {
		setDone();
		_abort = true;
		// Unterbrechung auslösen, sonst wird ArchiveBuffer ggf. nicht zeitnah beendet, wenn keine Daten eintreffen
		// (dann hängt er in winner.take(...) fest)
		this.interrupt();
		for(final ValueProvider vp : _realElements.values()) {
			vp.abort();
		}
	}

	/** @return Returns the aggregate. */
	protected boolean isAggregate() {
		return _aggregate;
	}

	/** @param aggregate The aggregate to set. */
	protected void setAggregate(final boolean aggregate) {
		this._aggregate = aggregate;
	}

	/**
	 * Ist direkt nachdem alle Daten einer Datenart eines Intevalls gesammelt wurden aufzurufen. Benachricht die Statusberechung und markiert den letzten
	 * Ausgangsdatensatz als den letzten des Intervalls.
	 *
	 * @param endOfInterval Endzeitpunkt des Intervalls, das gerade beendet wurde.
	 *
	 * @see IMDSBuilder#nextInterval(long)
	 */
	protected void nextInterval(final long endOfInterval) {
		_dataSetBuilder.nextInterval(endOfInterval);
	}

	/** @return <code>true</code> falls die Datenaufbereitung abgebrochen wurde. */
	public boolean isAbort() {
		return _abort;
	}

	/**
	 * Liefert das Attribut zurück, das den Zeitstempel des temporären Attributsfestlegt. Diese Methode liefert erst nach AUfruf der {@link #init()} Methode
	 * gültige Werte!
	 *
	 * @param tad Temporäres Attribut
	 *
	 * @return Reales Attribut, dessen Zeitstempel auch für das temporäre Attribut gilt. <code>null</code> wenn noch kein reales Attribut für das übergebene
	 *         temporäre Attribut eingetragen wurde.
	 */
	public RealAttributeDescription getTimeStampOrigin(final TempAttributeDescription tad) {
		if(_tempElements != null) {
			final TempAttribut ta = _tempElements.get(tad);
			if(ta != null) return ta.getTimeStampOrigin();
		}
		return null;
	}

	/**
	 * Zeigt an, ob ein Änderungsprotokoll erstellt wird.
	 *
	 * @return <code>true</code>falls ein Änderungsprotokoll erstellt wird. <code>false</code>, falls ein Zustandsprotokoll erstellt wird.
	 */
	public ProtocolType getProtocolType() {
		return _processingInformation.getProtocolType();
	}

	/**
	 * Zeigt an ob ein leerer Datensatz eingefügt werden soll.
	 *
	 * @return <code>true</code> falls ein leerer Datensatz eingefügt werden soll.
	 */
	protected boolean getInsertEmpty() {
		return _insertEmpty;
	}

	/**
	 * Liefert die Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @return Bezeichnung des Skripts
	 */
	public String getScriptName() {
		return _scriptName;
	}

}
